from libifstate.util import logger, ipr, LogStyle
from libifstate.exception import LinkTypeUnknown
from abc import ABC, abstractmethod

class Link(ABC):
    _nla_prefix = 'IFLA_'
    _classes = {}

    def __new__(cls, *args, **kwargs):
        cname = cls.__name__
        if cname == Link.__name__:
            cname = "{}Link".format(kwargs["kind"].lower().capitalize())

        for c in Link.__subclasses__():
            if c.__name__ == cname:
                return super().__new__(c)

        return super().__new__(GenericLink)
        #raise LinkTypeUnknown()

    def __init__(self, name, **kwargs):
        self.cap_create = True
        self.settings = {
            'ifname': name,
        }
        self.settings.update(kwargs)
        self.attr_map = {
            'kind': ['IFLA_LINKINFO', 'IFLA_INFO_KIND'],
        }
        self.attr_idx = ['link']
        self.idx = None

        if 'address' in self.settings:
            self.idx = next(iter(ipr.link_lookup(
                address=self.settings['address'])), None)
        if self.idx is None:
            self.idx = next(iter(ipr.link_lookup(ifname=name)), None)

    def _drill_attr(self, data, keys):
        key = keys[0]
        d = data.get_attr(key)

        if d is not None:
            if len(keys) > 1:
                return self._drill_attr(d, keys[1:])
            else:
                return d
        else:
            return None

    def get_if_attr(self, key):
        if key == "state":
            return self.iface['state']

        if key in self.attr_map:
            return self._drill_attr(self.iface, self.attr_map[key])

        nla = self.name2nla(key)
        ret = self.iface.get_attr(nla)
        if not ret is None:
            return ret

        info = self.iface.get_attr('IFLA_LINKINFO')
        if not info is None:
            ret = info.get_attr(nla)
            if not ret is None:
                return ret

            info = info.get_attr('IFLA_INFO_DATA')
            if not info is None:
                ret = info.get_attr(nla)
                if not ret is None:
                    return ret

        return None

    def apply(self):
        # lookup for attributes requiring a interface index
        for attr in self.attr_idx:
            if attr in self.settings:
                self.settings[attr] = next(iter(ipr.link_lookup(ifname=self.settings[attr])), self.settings[attr])

        if self.idx is not None:
            self.iface = next(iter(ipr.get_links(self.idx)), None)

            if self.cap_create and self.get_if_attr('kind') != self.settings['kind']:
                self.recreate()
            else:
                self.update()
        else:
            self.create()

    def create(self, oper="add"):
        logger.info(oper, extra={'iface': self.settings['ifname'], 'style': LogStyle.CHG})

        logger.debug("ip link add: {}".format( " ".join("{}={}".format(k, v) for k,v in self.settings.items()) ))
        ipr.link('add', **(self.settings))

    def recreate(self):
        logger.debug('has wrong link kind %s, removing', self.settings['kind'], extra={'iface': self.settings['ifname']})
        ipr.link('del', index=self.idx)
        self.idx = None
        self.create("replace")

    def update(self):
        logger.debug('checking', extra={'iface': self.settings['ifname']})

        old_state = self.iface['state']
        has_changes = False
        for setting in self.settings.keys():
            logger.debug('  %s: %s => %s', setting, self.get_if_attr(
                setting), self.settings[setting], extra={'iface': self.settings['ifname']})
            if setting != 'kind' or self.cap_create:
                has_changes |= self.get_if_attr(setting) != self.settings[setting]

        if has_changes:
            logger.debug('needs to be configured', extra={'iface': self.settings['ifname']})
            if old_state:
                logger.debug('shutting down', extra={'iface': self.settings['ifname']})
                ipr.link('set', index=self.idx, state='down')
                if not 'state' in self.settings:
                    self.settings['state'] = 'up'

            logger.info('change', extra={'iface': self.settings['ifname'], 'style': LogStyle.CHG})
            ipr.link('set', index=self.idx, **(self.settings))
        else:
            logger.info('ok', extra={'iface': self.settings['ifname'], 'style': LogStyle.OK})

    def depends(self):
        return None

    @classmethod
    def name2nla(self, name):
        '''
        Convert human-friendly name into NLA name

        Example: address -> IFLA_ADDRESS

        Requires self.prefix to be set
        '''
        name = name.upper()
        if name.find(self._nla_prefix) == -1:
            name = "%s%s" % (self._nla_prefix, name)
        return name


class GenericLink(Link):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import logging
from pyroute2 import IPRoute

logger = logging.getLogger('ifstate')
ipr = IPRoute()

class LogStyle():
    OK = "ok"
    CHG = "chg"
    DEL = "del"

    def colorize(style):
        if style == LogStyle.OK:
            return AnsiColors.GREEN
        if style == LogStyle.CHG:
            return AnsiColors.YELLOW
        if style == LogStyle.DEL:
            return AnsiColors.YELLOW
        return ""

class AnsiColors():
    GREEN = "\033[32m"
    RED = "\033[31m"
    MAGENTA = "\033[35m"
    YELLOW = "\033[33m"
    RESET = "\033[0m"
    BOLD = "\033[1m"



from libifstate.exception import LinkDuplicate
from libifstate.link.base import Link
from libifstate.address import Addresses
from libifstate.routing import Tables, Rules
from libifstate.parser import Parser
from libifstate.util import logger, ipr, LogStyle
from libifstate.exception import LinkCircularLinked, LinkNoConfigFound
from ipaddress import ip_network, ip_interface
import re

__version__ = "0.6.0"

class IfState():
    def __init__(self):
        self.links = {}
        self.addresses = {}
        self.ignore = {}
        self.tables = None
        self.rules = None
    
    def update(self, ifstates):
        # add interfaces from config
        for ifstate in ifstates['interfaces']:
            name = ifstate['name']
            if name in self.links:
                raise LinkDuplicate()
            if 'link' in ifstate:
                self.links[name] = Link(name, **ifstate['link'])
            else:
                self.links[name] = None

            if 'addresses' in ifstate:
                self.addresses[name] = Addresses(name, ifstate['addresses'])
            else:
                self.addresses[name] = None

        # add routing from config
        if 'routing' in ifstates:
            if 'routes' in ifstates['routing']:
                if self.tables is None:
                    self.tables = Tables()
                for route in ifstates['routing']['routes']:
                    self.tables.add(route)

            if 'rules' in ifstates['routing']:
                if self.rules is None:
                    self.rules = Rules()
                for rule in ifstates['routing']['rules']:
                    self.rules.add(rule)

        # add ignore list items
        self.ignore.update(ifstates['ignore'])

    def apply(self):
        self.ipaddr_ignore = set()
        for ip in self.ignore.get('ipaddr', []):
            self.ipaddr_ignore.add( ip_network(ip) )

        if not any(not x is None for x in self.links.values()):
            logger.error("DANGER: Not a single link config has been found!")
            raise LinkNoConfigFound()

        logger.info('configuring interface links')

        applied = []
        while len(applied) < len(self.links):
            last = len(applied)
            for name, link in self.links.items():
                if link is None:
                    logger.debug('skipped due to no link settings', extra={'iface': name})
                    applied.append(name)
                else:
                    dep = link.depends()
                    if dep is None or dep in applied:
                        link.apply()
                        applied.append(name)
            if last == len(applied):
                raise LinkCircularLinked()

        for link in ipr.get_links():
            name = link.get_attr('IFLA_IFNAME')
            # skip links on ignore list
            if not name in self.links and not any(re.match(regex, name) for regex in self.ignore.get('ifname', [])):
                info = link.get_attr('IFLA_LINKINFO')
                # remove virtual interface
                if info is not None:
                    kind = info.get_attr('IFLA_INFO_KIND')
                    logger.info('del', extra={'iface': name, 'style': LogStyle.DEL})
                    ipr.link('set', index=link.get('index'), state='down')
                    ipr.link('del', index=link.get('index'))
                # shutdown physical interfaces
                else:
                    if link.get('state') == 'down':
                        logger.warning('orphan', extra={'iface': name, 'style': LogStyle.OK})
                    else:
                        logger.warning('orphan', extra={'iface': name, 'style': LogStyle.CHG})
                        ipr.link('set', index=link.get('index'), state='down')

        if any(not x is None for x in self.addresses.values()):
            logger.info("\nconfiguring interface ip addresses...")
            # add empty objects for unhandled interfaces
            for link in ipr.get_links():
                name = link.get_attr('IFLA_IFNAME')
                # skip links on ignore list
                if not name in self.addresses and not any(re.match(regex, name) for regex in self.ignore.get('ifname', [])):
                    self.addresses[name] = Addresses(name, [])

            for name, addresses in self.addresses.items():
                if addresses is None:
                    logger.debug('skipped due to no address settings', extra={'iface': name})
                else:
                    addresses.apply(self.ipaddr_ignore)
        else:
            logger.info("\nno interface ip addressing to be applied")

        if not self.tables is None:
            self.tables.apply(self.ignore.get('routes', []))

        if not self.rules is None:
            self.rules.apply(self.ignore.get('rules', []))

    def show(self):
        self.ipaddr_ignore = set()
        for ip in Parser._default_ifstates.get('ignore').get('ipaddr'):
            self.ipaddr_ignore.add( ip_network(ip) )

        ifs_links = []
        for ipr_link in ipr.get_links():
            name = ipr_link.get_attr('IFLA_IFNAME')
            # skip links on ignore list
            if not any(re.match(regex, name) for regex in Parser._default_ifstates['ignore'].get('ifname', [])):
                ifs_link = {
                        'name': name,
                        'addresses': [],
                        'link': {
                            'state': ipr_link['state'],
                        },
                }

                for addr in ipr.get_addr(index=ipr_link['index']):
                    ip = ip_interface(addr.get_attr('IFA_ADDRESS') + '/' + str(addr['prefixlen']))
                    if not any(ip in net for net in self.ipaddr_ignore):
                        ifs_link['addresses'].append(ip.with_prefixlen)

                info = ipr_link.get_attr('IFLA_LINKINFO')
                if info is not None:
                    kind = info.get_attr('IFLA_INFO_KIND')
                    ifs_link['link']['kind'] = kind

                    data = info.get_attr('IFLA_INFO_DATA')
                    # unsupported link type, fallback to raw encoding
                    if type(data) == str:
                        ifs_link['link']["info_data"] = data
                    elif data is not None:
                        for k, v in data['attrs']:
                            ifs_link['link'][ipr_link.nla2name(k)] = v
                else:
                    ifs_link['link']['kind'] = 'physical'
                    addr = ipr_link.get_attr('IFLA_ADDRESS')
                    if not addr is None:
                        ifs_link['link']['address'] = addr

                ifs_links.append(ifs_link)

        routing = {
            'routes': Tables().show_routes(Parser._default_ifstates['ignore']['routes']),
        }

        return { **Parser._default_ifstates, **{'interfaces': ifs_links, 'routing': routing}}



class LinkCannotAdd(Exception):
    pass

class LinkTypeUnknown(Exception):
    pass

class LinkDuplicate(Exception):
    pass

class LinkCircularLinked(Exception):
    pass

class LinkNoConfigFound(Exception):
    pass

class RouteDupblicate(Exception):
    pass



from libifstate.util import logger, ipr, LogStyle
from libifstate.exception import RouteDupblicate
from ipaddress import ip_address, ip_network
from pyroute2.netlink.exceptions import NetlinkError
import collections.abc
from glob import glob
import os
import re
import sys
from socket import AF_INET, AF_INET6

def route_matches(r1, r2, fields=('dst', 'metric', 'proto'), verbose=False):
    for fld in fields:
        if verbose:
            logger.debug("{}: {} - {}".format(fld, r1.get(fld), r2.get(fld)))
        if fld in r1 and fld in r2:
            if r1[fld] != r2[fld]:
                return False
        elif fld in r1 or fld in r2:
                return False
    return True

def route_identical(r1, r2):
    return route_matches(r1, r2, set(r1.keys()) | set(r2.keys()), True)

class RTLookup():
    def __init__(self, name):
        self.name = name
        self.str2id = {}
        self.id2str = {}

        try:
            fn = os.path.join('/etc/iproute2', name)
            with open(fn, 'r') as fp:
                self._parse(fp)

            for fn in glob(os.path.join('/etc/iproute2', "{}.d".format(name), "*.conf")):
                with open(fn, 'r') as fp:
                    self._parse(fp)
        except:
            logger.info('could not open {}'.format(fn))

    def _parse(self, fp):
        for line in fp:
            m = re.search(r'^(\d+)\s+(\S+)$', line)
            if m:
                self.str2id[m.group(2)] = int(m.group(1))
                self.id2str[int(m.group(1))] = m.group(2)

    def lookup_id(self, key):
        if type(key) == int or key.isdecimal():
            return int(key)
        
        return self.str2id[key]

    def lookup_str(self, key):
        return self.id2str.get(key, key)

class RTLookups():
    tables = RTLookup('rt_tables')
    realms = RTLookup('rt_realms')
    scopes = RTLookup('rt_scopes')
    protos = RTLookup('rt_protos')

class Tables(collections.abc.Mapping):
    def __init__(self):
        self.tables = {
            254: [],
        }

    def __getitem__(self, key):
        if not key in self.tables:
            raise KeyError()

        return self.tables[key]

    def __iter__(self):
        return self.tables.__iter__()

    def __len__(self):
        return len(self.tables)

    def add(self, route):
        rt = {
            'table': RTLookups.tables.lookup_id( route.get('table', 254) ),
            'dst': ip_network(route['to']).with_prefixlen,
            'scope': RTLookups.scopes.lookup_id(route.get('scope', 0)),
            'proto': RTLookups.protos.lookup_id(route.get('proto', 3)),
            'realm': RTLookups.realms.lookup_id(route.get('realm', 0)),
            'tos': route.get('tos', 0),
        }

        if 'dev' in route:
            rt['oif'] = next(iter(ipr.link_lookup(ifname=route['dev'])))

        if 'via' in route:
            rt['gateway'] = str(ip_address(route['via']))

        if not rt['table'] in self.tables:
            self.tables[rt['table']] = []
        self.tables[rt['table']].append(rt)

    def show_routes(self, ignores):
        routes = []
        for route in ipr.get_routes(family=AF_INET) + ipr.get_routes(family=AF_INET6):
            # skip routes from local table
            table = route.get_attr('RTA_TABLE')
            if table == 255:
                continue

            # skip ignored routes
            if route['proto'] in ignores.get('protos', []):
                continue

            if route['dst_len'] > 0:
                dst = ip_network('{}/{}'.format(route.get_attr('RTA_DST'), route['dst_len'])).with_prefixlen
            elif route['family']==AF_INET:
                dst = ip_network('0.0.0.0/0').with_prefixlen
            elif route['family']==AF_INET6:
                dst = ip_network('::/0').with_prefixlen

            rt = {
                'to': dst,
                'table': RTLookups.tables.lookup_str(table),
            }

            dev = route.get_attr('RTA_OIF')
            if dev:
                link = next(iter(ipr.get_links(dev)), None)
                if link:
                    rt['dev'] = link.get_attr('IFLA_IFNAME', dev)
                else:
                    rt['dev'] = dev

            realm = route.get_attr('RTA_FLOW')
            if realm:
                rt['realm'] = RTLookups.realms.lookup_str(realm)

            if route['scope'] != 0:
                rt['scope'] = RTLookups.scopes.lookup_str(route['scope'])

            if route['proto'] != 3:
                rt['proto'] = RTLookups.protos.lookup_str(route['proto'])

            if route['tos'] != 0:
                rt['tos'] = route['tos']

            routes.append(rt)

        return routes

    def kernel_routes(self, table):
        routes = []
        for route in ipr.get_routes(table=table, family=AF_INET) + ipr.get_routes(table=table, family=AF_INET6):
            if route['dst_len'] > 0:
                dst = ip_network('{}/{}'.format(route.get_attr('RTA_DST'), route['dst_len'])).with_prefixlen
            elif route['family']==AF_INET:
                dst = ip_network('0.0.0.0/0').with_prefixlen
            elif route['family']==AF_INET6:
                dst = ip_network('::/0').with_prefixlen

            rt = {
                'table': route.get_attr('RTA_TABLE'),
                'dst': dst,
                'oif': route.get_attr('RTA_OIF'),
                'scope': route['scope'],
                'proto': route['proto'],
                'realm': route.get_attr('RTA_FLOW', 0),
                'tos': route['tos'],
            }

            gateway = route.get_attr('RTA_GATEWAY')
            if not gateway is None:
                rt['gateway'] = str(ip_address(gateway))

            metric = route.get_attr('RTA_PRIORITY')
            if not metric is None:
                rt['metric'] = metric

            pref = route.get_attr('RTA_PREF')
            if not pref is None:
                rt['pref'] = pref

            routes.append(rt)
        return routes

    def apply(self, ignores):
        for table, croutes in self.tables.items():
            pfx = RTLookups.tables.lookup_str(table)
            logger.info('\nconfiguring routing table {}...'.format(pfx))

            kroutes = self.kernel_routes(table)

            for route in croutes:
                found = False
                identical = False
                for i, kroute in enumerate(kroutes):
                    if route_matches(route, kroute):
                        del kroutes[i]
                        found = True
                        if route_identical(route, kroute):
                            identical = True
                            break

                if identical:
                    logger.info('ok', extra={'iface': route['dst'], 'style': LogStyle.OK})
                else:
                    if found:
                        logger.info('change', extra={'iface': route['dst'], 'style': LogStyle.CHG})
                    else:
                        logger.info('add', extra={'iface': route['dst'], 'style': LogStyle.CHG})

                    logger.debug("ip route replace: {}".format( " ".join("{}={}".format(k, v) for k,v in route.items()) ))
                    try:
                        ipr.route('replace', **route)
                    except NetlinkError as err:
                        logger.warning('setup route {} failed: {}'.format(route['dst'], err.args[1]))

            for route in kroutes:
                if route['proto'] in ignores.get('protos', []):
                    continue

                logger.info('del', extra={'iface': route['dst'], 'style': LogStyle.DEL})
                ipr.route('del', **route)


class Routes():
    def __init__(self):
        self.routes = {}

    def add(self, route):
        dst = ip_network(route['dst'])
        if dst in self.routes:
            raise RouteDupblicate()

        if 'table' in route:
            route['table'] = RTLookups.tables.lookup_id(route.get_attr('RTA_TABLE', 'main'))
        # if 'dev' in route:
        #     route['dev'] = xxxx
        if 'scope' in route:
            route['scope'] = RTLookups.scopes.lookup_id(route['scope'])

        if 'proto' in route:
            route['proto'] = RTLookups.protos.lookup_id(route['proto'])

        if 'gateway' in route:
            route['gateway'] = ip_address(route['gateway'])

        if 'realm' in route:
            route['realm'] = RTLookups.realms.lookup_id(route['realm'])

        self.routes[dst] = route


class Rules():

    def add(self, rule):
        pass

    def apply(self, ignores):
        pass



from setuptools import setup, find_packages
from datetime import date


def readme():
    with open("README.md") as f:
        return f.read()


def requirements():
    req = []
    with open("requirements.txt") as fd:
        for line in fd:
            line.strip()
            if not line.startswith("#"):
                req.append(line)
    return req

def version():
    with open("libifstate/__init__.py") as fd:
        for line in fd:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

setup(
    name="ifstate",
    version=version(),
    description="Manage host interface settings in a declarative manner",
    author="Thomas Liske",
    author_email="thomas@fiasko-nw.net",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/liske/ifstate",
    license="GPL3+",
    packages=find_packages(),
    install_requires=[
        "pyroute2",
        "PyYAML",
    ],
    entry_points={
        "console_scripts": ["ifstatecli = ifstate.ifstate:main"]
    },
)



from libifstate.util import logger, ipr, LogStyle
from ipaddress import ip_interface
#from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg

class Addresses():
    def __init__(self, iface, addresses):
        self.iface = iface
        self.addresses = []
        for address in addresses:
            self.addresses.append(ip_interface(address))

    def apply(self, ignore):
        logger.debug('getting addresses', extra={'iface': self.iface})

        # get ifindex
        idx = next(iter(ipr.link_lookup(ifname=self.iface)))

        # get active ip addresses
        ipr_addr = {}
        for addr in ipr.get_addr(index=idx):
            ip = ip_interface(addr.get_attr('IFA_ADDRESS') + '/' + str(addr['prefixlen']))
            ipr_addr[ip] = addr
#            ipr_flags[ip] = ifaddrmsg.flags2names(addr['flags'], family=addr['family'])

        for addr in self.addresses:
            ip = str(addr.ip)
            if addr in ipr_addr:
                logger.info('%s', addr.with_prefixlen, extra={'iface': self.iface, 'style': LogStyle.OK})
                del ipr_addr[addr]
            else:
                logger.info('%s', addr.with_prefixlen, extra={'iface': self.iface, 'style': LogStyle.CHG})
                ipr.addr("add", index=idx, address=ip, mask=addr.network.prefixlen)

        for ip, addr in ipr_addr.items():
            if not any(ip in net for net in ignore):
#                if 'IFA_F_PERMANENT' in ipr_flags[ip]:
                logger.info('%s', ip.with_prefixlen, extra={'iface': self.iface, 'style': LogStyle.DEL})
                ipr.addr("del", index=idx, address=str(ip.ip), mask=ip.network.prefixlen)



#!/usr/bin/env python3

from libifstate.parser import YamlParser
from libifstate import __version__, IfState
from libifstate.util import logger, LogStyle , AnsiColors
from collections import namedtuple

import argparse
import logging
import signal
import sys
import yaml

class Actions():
    CHECK = "check"
    APPLY = "apply"
    SHOW = "show"

class LogFilter(logging.Filter):
    def __init__(self, terminal):
        super().__init__()
        self.terminal = terminal

    def filter(self, record):
        record.levelshort = record.levelname[:1]
        if hasattr(record, 'iface'):
            record.prefix = " {:15} ".format(record.iface)
        else:
            record.prefix = ''

        if self.terminal:
            if record.levelno >= logging.ERROR:
                record.bol = AnsiColors.RED
            elif record.levelno >= logging.WARNING:
                record.bol = AnsiColors.MAGENTA
            else:
                record.bol = ''
            record.eol = AnsiColors.RESET
        else:
            record.bol = ''
            record.eol = ''

        if hasattr(record, 'style'):
            if self.terminal:
                record.style = LogStyle.colorize(record.style)
            else:
                record.style = ""
        else:
            record.style = ""

        return True

def setup_logging(level):
    if level != logging.DEBUG:
        sys.tracebacklimit = 0

    logging.basicConfig(
        level=level,
        format='%(bol)s%(prefix)s%(style)s%(message)s%(eol)s',
    )

    f = LogFilter(sys.stderr.isatty())
    logger.addFilter(f)

def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-q", "--quiet", action="store_true", help="be more quiet, print only warnings and errors")
    group.add_argument("-v", "--verbose", action="store_true", help="be more verbose")
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument("-c", "--config", type=str, default="/etc/ifstate/config.yml", help="configuration YaML filename")
    parser.add_argument("action", choices=list(a.lower() for a in dir(Actions) if not a.startswith('_')), help="specifies the action to perform")

    args = parser.parse_args()
    if args.verbose:
        setup_logging(logging.DEBUG)
    elif args.quiet:
        setup_logging(logging.ERROR)
    else:
        setup_logging(logging.INFO)

    ifs = IfState()

    if args.action == Actions.SHOW:
        # preserve dict order on python 3.7+
        if sys.version_info >= (3,7):
            yaml.add_representer(dict, lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()))
        print(yaml.dump(ifs.show()))

    if args.action in [Actions.CHECK, Actions.APPLY]:
        parser = YamlParser(args.config)
        ifstates = parser.config()

        ifs.update(ifstates)

        if args.action == Actions.CHECK:
            pass
        else:
            # ignore some well-known signals to prevent interruptions (i.e. due to ssh connection loss)
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            ifs.apply()

if __name__ == "__main__":
    main()
